package highscoretetrisai.real.bot;

import java.util.*;
import nintaco.api.*;
import highscoretetrisai.real.*;
import static highscoretetrisai.real.Playfield.*;

public class TetrisBot {
  
  private static final int EMPTY_SQUARE = 0xEF;
  
  private final API api = ApiSource.getAPI();  
  
  private final boolean[][] playfield = createPlayfield();
  private final SearchChain searchChain = new SearchChain();
  
  private final List<Coordinate> moves = new ArrayList<>();
  private final int[] TetriminosTypes = new int[19];

  private int currentType;
  private int nextType;

  private int lastPlayState;
  private int startCounter;
  private int movesIndex;
  private boolean moving;  
  
  public TetrisBot() {
    searchChain.setWeights(new double[] { 
      0.2861270952978939, 
      1.7012336769099592, 
      0.7113042307683077, 
      0.9106654159986804, 
      1.879338064244357, 
      2.168463848297177, 
      -0.26558711196175727, 
      0.2898865849496105, 
      0.36236105526118173, 
      -0.028668795795469625, 
      0.8741799811132331, 
      -0.5074096831443619, 
      -2.148676202831281, 
      -1.1875585402811417, 
      -2.645656132241128, 
      0.24204341626870662, 
      0.28783812616443144, 
    });
  }
  
  public void launch() throws Throwable {
    api.addActivateListener(this::apiEnabled);
    api.addAccessPointListener(this::updateScore, AccessPointType.PreExecute, 
        0x9C35);
    api.addFrameListener(this::renderFinished);
    api.addStatusListener(this::statusChanged);
    api.run();
  }
  
  private void apiEnabled() {
    readTetriminoTypes();
  }  
  
  private void readTetriminoTypes() {
    for(int i = 0; i < 19; i++) {
      TetriminosTypes[i] = api.readCPU(Addresses.TetriminoTypeTable + i);
    }
  }  
  
  private void resetPlayState(final int gameState) {
    if (gameState != 4) {
      api.writeCPU(Addresses.PlayState, 0);
    }
  }
  
  private int updateScore(final int type, final int address, 
      final int value) {
    // cap the points multiplier at 30 to avoid the kill screen
    if (api.readCPU(0x00A8) > 30) {
      api.writeCPU(0x00A8, 30);
    }
    return -1;
  }

  private int readTetrimino() {
    return TetriminosTypes[api.readCPU(Addresses.TetriminoID)];
  }

  private int readNextTetrimino() {
    return TetriminosTypes[api.readCPU(Addresses.NextTetriminoID)];
  }  
  
  private void readPlayfield() {
    currentType = readTetrimino();
    nextType = readNextTetrimino();
    
    for(int y = HEIGHT - 1; y >= 0; --y) {
      final int address = Addresses.Playfield + 10 * y;
      final boolean[] row = playfield[y];
      for(int x = WIDTH - 1; x >= 0; --x) {
        row[x] = (api.readCPU(address + x) != EMPTY_SQUARE);
      }
    }    
  }
  
  private boolean spawned() {
    final int currentTetrimino = api.readCPU(Addresses.TetriminoID);
    final int playState = api.readCPU(Addresses.PlayState);
    final int tetriminoX = api.readCPU(Addresses.TetriminoX);
    final int tetriminoY = api.readCPU(Addresses.TetriminoY1);
    
    final boolean result = lastPlayState != 1 && playState == 1 
        && tetriminoX == 5 && tetriminoY == 0 
            && currentTetrimino < TetriminosTypes.length;
    lastPlayState = playState;
    return result;
  }
  
  private boolean isPlaying(final int gameState) {
    return gameState == 4 && api.readCPU(Addresses.PlayState) < 9;
  }  
  
  private void pressStart() {
    if (startCounter > 0) {
      startCounter--;
    } else {
      startCounter = 10;
    } 
    if (startCounter >= 5) {
      api.writeGamepad(0, GamepadButtons.Start, true);
    }    
  }  
  
  private void skipCopyrightScreen(final int gameState) {
    if (gameState == 0) {
      if (api.readCPU(Addresses.Copyright1) > 1) {
        api.writeCPU(Addresses.Copyright1, 0);
      } else if (api.readCPU(Addresses.Copyright2) > 2) {
        api.writeCPU(Addresses.Copyright2, 1);
      }
    }
  }
  
  private void skipTitleAndDemoScreens(final int gameState) {
    if (gameState == 1 || gameState == 5) {
      pressStart();   
    } else {
      startCounter = 0;
    }
  }
  
  private void makeMove(final Coordinate coordinate) {
    api.writeGamepad(0, GamepadButtons.A, coordinate.A);
    api.writeGamepad(0, GamepadButtons.B, coordinate.B);
    api.writeGamepad(0, GamepadButtons.Left, coordinate.left);
    api.writeGamepad(0, GamepadButtons.Right, coordinate.right);
  }
  
  private void renderFinished() {
    final int gameState = api.readCPU(Addresses.GameState);
    skipCopyrightScreen(gameState);
    skipTitleAndDemoScreens(gameState);
    resetPlayState(gameState);
    
    if (isPlaying(gameState)) {
      api.writeGamepad(0, GamepadButtons.A, false);
      final boolean level29 = (api.readCPU(Addresses.Level) >= 29);
      if (spawned()) {        
        readPlayfield();
        searchChain.setLevel29(level29);
        searchChain.search(currentType, nextType, playfield);          
        if (searchChain.isBestFound()) {
          searchChain.getMoves(moves);          
          movesIndex = 0;
          moving = true;
        } else {
          moving = false;
        }
      }
            
      if (moving) {
        if (movesIndex < moves.size()) {
          makeMove(moves.get(movesIndex));
          if (level29) {
            movesIndex += 2;
          } else {
            ++movesIndex;
          }
        } else {
          api.writeGamepad(0, GamepadButtons.Down, true);
        }
      } 
    } else {
      moves.clear();
      moving = false;
    } 
  }
  
  private void statusChanged(final String message) {
    System.out.println(message);
  }
  
  public static void main(final String... args) throws Throwable {
    ApiSource.initRemoteAPI("localhost", 9999);    
    new TetrisBot().launch();
  }
}